/*==================================================================
  spectrum.cpp
  a cross-platform emulator for the Casio Prizm, fx-cg10
  incomplete: version 0.00.01
  please view readme before running
  written by z80man
  ==================================================================*/
  
#include <iostream.h>
#include <time.h>
#include <fstream.h>
#include <stdio.h>

void display_reg(int speed);
void display_reg_t(int speed);
int readmemory_long(int pointer);
void writememory_long(int pointer, int data);
short readmemory_word(int pointer);
void writememory_byte(int pointer, char data);
void ADD();
void MOVLi();
void MOVr();
void MOVLim();
void MOVin();
void MOVLpush();
void MOVLpop();
void CMPEQr();
void DT();
void SUB();
void ANDr();
void NOT();
void ORr();
void TSTr();
void XORr();
void BF();
void BT();
void NOP();
bool decode();
unsigned int PC;
__inline unsigned short htons(unsigned short n)
{
  return ((n & 0xFF) << 8) | ((n & 0xFF00) >> 8);
}

__inline unsigned short ntohs(unsigned short n)
{
  return ((n & 0xFF) << 8) | ((n & 0xFF00) >> 8);
}

__inline unsigned long htonl(unsigned long n)
{
  return ((n & 0xFF) << 24) | ((n & 0xFF00) << 8) | ((n & 0xFF0000) >> 8) | ((n & 0xFF000000) >> 24);
}

__inline unsigned long ntohl(unsigned long n)
{
  return ((n & 0xFF) << 24) | ((n & 0xFF00) << 8) | ((n & 0xFF0000) >> 8) | ((n & 0xFF000000) >> 24);
}

/*===========================================
  instruction index.h
  used with spectrum.cpp
  defines opcodes for SH3 instructions
  written by z80man
  ===========================================*/

#define ADD_v       0x300C
#define MOVi_v      0xE000
#define ADDi_v      0x7000     
#define MOVLi_v     0xD000
#define MOVr_v      0x6003
#define MOVLin_v    0x2002
#define MOVLim_v    0x6002
#define MOVLpush_v  0x2006
#define MOVLpop_v   0x6006
#define CMPEQr_v    0x3000
#define DT_v        0x4010
#define SUB_v       0x3080
#define ANDr_v      0x2009
#define NOT_v       0x6007
#define ORr_v       0x200B
#define TSTr_v      0x2008
#define XORr_v      0x200A
#define BF_v        0x8B00
#define BT_v        0x8D00
#define NOP_v       0x0009
#define END_v       0xFFFF
#define PAUSE_v     0xFFFE


#define ADD_vv       0xF00F
#define MOVi_vv      0xF000
#define ADDi_vv      0xF000     
#define MOVLi_vv     0xF000
#define MOVr_vv      0xF00F
#define MOVLin_vv    0xF00F
#define MOVLim_vv    0xF00F
#define MOVLpush_vv  0xF00F
#define MOVLpop_vv   0xF00F
#define CMPEQr_vv    0xF000
#define DT_vv        0xF0F0
#define SUB_vv       0xF0F0
#define ANDr_vv      0xF00F
#define NOT_vv       0xF00F
#define ORr_vv       0xF00F
#define TSTr_vv      0xF00F
#define XORr_vv      0xF00F
#define BF_vv        0xFF00
#define BT_vv        0xFF00
#define NOP_vv       0x000F
#define END_vv       0xFFFF
#define PAUSE_vv     0xFFFF	

/*==============================================================
  global variables.h
  used with spectrum.cpp
  intialization of general purpose, system, and control registers
  written by z80man
  ==============================================================*/

unsigned int R[16];      //general purpose register array R0-R15
unsigned int R_banked[8]; //banked register array R0_BANK-R7_BANK
unsigned int SR,GBR,VBR,MACH,MACL,PR,SSR,SPC;
struct SR0 // bit struct for system register
{
	unsigned int dummy0:4;
	unsigned int RC0:12;
	unsigned int dummy1:4;
	unsigned int DMY0:1;
	unsigned int DMX0:1;
	unsigned int M0:1;
	unsigned int Q0:1;
	unsigned int I0:4;
	unsigned int RF10:1;
	unsigned int RF00:1;
	unsigned int S0:1;
	unsigned int T0:1;
};
struct savedSR0 // bit struct for saved system register
{
	unsigned int saveddummy0:4;
	unsigned int savedRC0:12;
	unsigned int saveddummy1:4;
	unsigned int savedDMY0:1;
	unsigned int savedDMX0:1;
	unsigned int savedM0:1;
	unsigned int savedQ0:1;
	unsigned int savedI0:4;
	unsigned int savedRF10:1;
	unsigned int savedRF00:1;
	unsigned int savedS0:1;
	unsigned int savedT0:1;
};
#define M ((*(struct SR0 *)(&SR)).M0) //defintion for SR0
#define Q ((*(struct SR0 *)(&SR)).Q0)
#define S ((*(struct SR0 *)(&SR)).S0)
#define T ((*(struct SR0 *)(&SR)).T0)
#define RF1 ((*(struct SR0 *)(&SR)).RF10)
#define RF0 ((*(struct SR0 *)(&SR)).RF00)
unsigned short instruction;
char * memory = new char[0xFFFF];

  
  
  int main(void)
{
	PC = 0xA4000000;
	ifstream program("spectrum.data", ios::in|ios::binary);
	if(!program.is_open())
	{
		cout << "Error opening spectrum.data";
		cin.get();
		return -1;
	}
	program.seekg(0, ios::beg);
	int program_counter;
	char buffer;
	program.read(&buffer, 1);
	int prog_index = 0; 
	while(program.good())
	{
		writememory_byte(prog_index + 0xA4000000, buffer);
		prog_index++;
		program.seekg(prog_index);
		program.read(&buffer, 1);
	}
	program.close();
	bool end = true;
	int speed;
	int tempspeed = time(0);
	while (1)
	{
		end = decode();
		speed++;
		if (tempspeed != time(0))
		{
			display_reg_t(speed);
			tempspeed = time(0);
			speed = 0;
			
		}
	}
	return 0;
}

/*==================================================================
  reg display.h
  used with spectrum.cpp
  outputs contents of registers R0-R15
  written by z80man
  ==================================================================*/


void display_reg_t(int speed)
{
	cout << "\nCurrent Speed: " << speed << " Hz\n";
	for (int counter = 0; counter <16; counter++)
	{
		cout << "\nR" << counter << " " << R[counter];
	}
	cout << endl;
}

	
		
/*==================================================================
  prizm memory.h
  used with spectrim.cpp
  conatins functions for accessing the memory
  written by z80man
=====================================================================*/




int readmemory_long(int pointer)
{
	unsigned int r_pointer = pointer - 0xA4000000;
	if (r_pointer > 0x1FFFFF) return 0;
	int temp = int(memory[r_pointer]) * 0x1000000;
	temp |= int(memory[r_pointer + 1]) * 0x10000;
	temp |= int(memory[r_pointer + 2]) * 0x100;
	temp |= int(memory[r_pointer + 3]);
	return htonl(temp);
	
}
void writememory_long(int pointer, int data)
{
	unsigned int r_pointer = pointer - 0xA4000000;
	if (r_pointer > 0x1FFFFF) return;
	memory[r_pointer + 3] = char(data / 0x1000000);
	memory[r_pointer + 2] = char((data & 0x00FF0000) / 0x10000);
	memory[r_pointer + 1] = char((data & 0x0000FF00) / 0x100);
	memory[r_pointer] = char(data & 0x000000FF);
}
short readmemory_word(int pointer)
{
	unsigned int r_pointer = pointer - 0xA4000000;
	if (r_pointer > 0x1FFFFF) return 0;
	short temp = short(memory[r_pointer]);
	temp |= short(memory[r_pointer + 1]) * 0x100;
	return htons(temp);
}
void writememory_byte(int pointer, char data)
{
	unsigned int r_pointer = pointer - 0xA4000000;
	if (r_pointer > 0x1FFFFF) return;
	memory[r_pointer] = data;	
}

		
				
/*=============================================
  instruction set.h
  used with spectrum.cpp
  contains function for each SH3 instruction
  written by z80man
  =============================================*/

void ADD() //ADD Rm,Rn
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] += R[m];
	PC += 2;
	return;
}
void MOVi()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int i = instruction & 0x000000FF;
	R[n] = i;
	PC += 2;
}
void ADDi()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int i = instruction & 0x000000FF;
	R[n] += i;
	PC += 2;
}
void MOVLi() // MOV.L @(disp,PC),Rn
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int d = instruction & 0x000000FF;
	R[n] = readmemory_long((d * 4) + ((PC + 2) & 0x0000000C));
	PC += 2;
	return;
}
void MOVr()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] = R[m];
	PC += 2;
	return;
}
void MOVLim()
{ 
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] = readmemory_long(R[m]);
	PC += 2;
	return;
}
void MOVLin()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	writememory_long(R[n], R[m]);
	PC += 2;
	return;
}
void MOVLpush()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] -= 4;
	writememory_long(R[n], R[m]);
	PC += 2;
	return;
}
void MOVLpop()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] = readmemory_long(R[m]);
	R[m] += 4;
	PC += 2;
	return;
}
void CMPEQr()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	if (R[n] == R[m]) T = 1;
	else T = 0;
	PC += 2;
	return;
}
void DT()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	R[n]--;
	if (R[n] == 0) T = 1;
	else T = 0;
	PC += 2;
	return;
}
void SUB()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] -= R[m];
	PC +=2;
	return;
}
void ANDr()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] &= R[m];
	PC += 2;
	return;
}
void NOT()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] = ~ R[m];
	PC += 2;
	return;
}
void ORr()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] |= R[m];
	PC += 2;
	return;
}
void TSTr()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	if ((R[n] & R[m]) == 0) T = 1;
	else T = 0;
	PC += 2;
	return;
}
void XORr()
{
	unsigned int n = (instruction & 0x0F00) / 0x100;
	unsigned int m = (instruction & 0x00F0) / 0x10;
	R[n] ^= R[m];
	PC += 2;
	return;
}
void BF()
{
	short d = instruction & 0x00FF;
	if ((d&0x80)==0) d=(0x00FF & d);
	else d=(0xFF00 | d);
	if (T==0) PC=PC - 2 + int(d<<1);
	else PC+=2;
	return;
}
void BT()
{
	unsigned int d = instruction & 0x000000FF;
	if ((d&0x80)==0) d=(0x000000FF & (long)d);
	else d=(0xFFFFFF00 | (long)d);
	if (T==1) PC=PC - 2 + (d<<1);
	else PC+=2;
	return;
}
void NOP()
{
	PC += 2;
	return;
}
	
/*================================================================
  decode.h
  used with spectrum.cpp
  calls instructions from the current PC
  written by z80man
  ================================================================*/
 
  
bool decode(void)
{
	bool end;
	instruction = readmemory_word(PC);     //load instruction from memory
	
	if (ADD_v == (instruction & ADD_vv))  //adds Rn and Rm into Rn: ADD Rm,Rn
		ADD();
		 
	else if (MOVi_v == (instruction & MOVi_vv)) 
		MOVi();
		
	else if (ADDi_v == (instruction & ADDi_vv)) 
		ADDi();
		
	else if (MOVLi_v == (instruction & MOVLi_vv))  //copies data at disp*4 into Rn: MOV.L @(disp,PC),Rn
		MOVLi();
		 
	else if (MOVr_v == (instruction & MOVr_vv))  //copies data from Rm into Rn: MOV Rm,Rn
		MOVr();
		 
	else if (MOVLin_v == (instruction & MOVLin_vv))  //copies data from Rm into adress of Rn: MOV.L Rm,@Rn  
		MOVLin();
		 
	else if (MOVLim_v == (instruction & MOVLim_vv))  //copies data from adress of Rm into Rn: MOV.L @Rm,Rn
		MOVLim();
		 
	else if (MOVLpush_v == (instruction & MOVLpush_vv))  //copies data from Rm into adress Rn. Rn decrements by four: MOV.L RM,@-Rn
		MOVLpush();
		 
	else if (MOVLpop_v == (instruction & MOVLpop_vv))  //copies data from adress of Rm into Rn. Rm increments by four: MOV.L @Rm+,Rn
		MOVLpop();
		 
	else if (CMPEQr_v == (instruction & CMPEQr_vv))  //if Rm is equal to Rn, then T equals 1: CMP/EQ Rm,Rn
		CMPEQr();
		 
	else if (DT_v == (instruction & DT_vv))  //Rn decrements by 1. If Rn equals 0, then 1 equals T
		DT();
		 
	else if (SUB_v == (instruction & SUB_vv))  //subtracts Rm from Rn: SUB Rm,Rn
		SUB();
		 
	else if (ANDr_v == (instruction & ANDr_vv))  //binary and of Rm and Rn into Rn: AND Rm,Rn
		ANDr();
		 
	else if (NOT_v == (instruction & NOT_vv))  //inverse of Rm becomes Rn: NOT Rm,Rn
		NOT();
		 
	else if (ORr_v == (instruction & ORr_vv))  //binary or of Rm and Rn becomes Rn: Or Rm,Rn
		ORr();
		 
	else if (TSTr_v == (instruction &	TSTr_vv))  //binary and of Rm and Rn. If result equals 0, then 1 equals T
		TSTr();
		 
	else if (XORr_v == (instruction & XORr_vv))  //binary xor of Rm and Rn into Rn: XOR Rm,Rn
		XORr();
		 
	else if (BF_v == (instruction & BF_vv))  //if T equals zero, then disp*2 + PC equals PC: BF label
		BF();
		 
	else if (BT_v == (instruction & BT_vv))  //if T equals 1, then disp*2 plus PC equals PC: BT label
		BT();
		 
	else if (NOP_v == instruction)  //no operation
		NOP();
		 
	else if (instruction == END_v)                   //not actual instruction, signifies end of execution
		end = false;
		
	else if (instruction == PAUSE_v) 
		cin.get();
		 
	else NOP();
	
	return end;	
}







		
	
	
	
					
		
	
	
		
	

	

		
	
  